# Generated by Django 5.2.5 on 2025-09-22 07:43
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0018_remove_type_field"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="ipfabrictransformmap",
            options={"ordering": ("pk",)},
        ),
        migrations.AlterModelOptions(
            name="ipfabrictransformmapgroup",
            options={"ordering": ("pk",)},
        ),
    ]
